// This is Universal iPod Display Board firmware
// Project: Axxess Display Board
// Comment: Displays information on the LCD from the serial port - refer to Axxess Display Specification
// Files: 	DisplayModule.c - Main file
//			LCD_Driver.c & LCD_Driver.h

//			
// Compiler:HiTech PICC Compiler, V8.02 PL1
// Hardware:CM05018 
//
// Author:	Andy Beck
// Revs:	01- Initial Release

#define VER_STRING "<VER 01>"				// Defines version of this firmware

#include <pic.h>
#include <string.h>
#include "LCD_Driver.h"

#define SETTINGS_STRING		"V021604"		// Defines module configurations - 2 digits each for lines, chars/line, & buttons

#define TRUE 1
#define FALSE 0
#define Dim(x) (sizeof(x) / sizeof(x[0]))
#define	XTAL_FREQ	4		/* Crystal frequency in MHz */

// Define intervals that drive the auto repeat & timed key functions
#define REPEAT_DELAY  	 500		// Delay 1/2 Second before Auto-Repeating
#define REPEAT_INTERVAL	 30			// Auto Repeat every 30ms (33 keys/Sec)
#define DEBOUNCE_MSEC  	 10			// Interval (in ms) between calls to debounce function

// Define the count value for the auto repeat parameters
#define DELAY_COUNT 			(REPEAT_DELAY / DEBOUNCE_MSEC)	
#define RATE_COUNT				(REPEAT_INTERVAL / DEBOUNCE_MSEC)
#define TIMED_KEY_COUNT			(TIMED_KEY_TIME / DEBOUNCE_MSEC)
#define DEBOUNCE_CYCLES 5		// Number of times the same key data has to be returned to be valid (debounced) 
#define NO_KEY 	0				

static unsigned char keypad_data[DEBOUNCE_CYCLES];	// Contains the bit patterns from the last n keypad scans
static char key;	

// Some Serial I/O macros to make the code cleaner
#define SIO_REC_READY (RCIF==1)
#define SIO_GET (RCREG)
#define SIO_OUT_BUSY (TRMT==0)
#define SIO_OUT_EMPTY (TRMT==1)
#define SIO_SEND(x)	(TXREG = x)
#define SIO_XMIT(x) {while(SIO_OUT_BUSY); TXREG = x;}

#define MAX_CONTRAST 16								// Provide 16 bits of Contrast Control
#define TEXT_BUFFER_LENGTH 16
#define RECEIVE_BUFFER_LENGTH 32

#define STX_CHAR 0xFF								// STX character is first character in all message
#define ETX_CHAR 0x00								// ETX is last character in all message

#define ENTER_SWITCH		0x08					// Bit mask for Switch A (S1 = Enter)
#define ESC_SWITCH			0x10					// Bit mask for Switch B (S2 = ESC)
#define UP_SWITCH			0x40					// Bit mask for Switch C (S3 = Up)
#define DOWN_SWITCH			0x80					// Bit mask for Switch D (S4 = Down)
#define ENTER_ESC_SWITCH	0x18					// Switches A & B together
#define UP_DOWN_SWITCH   	0xC0					// Switches C & D together
#define ENTER_UP_SWITCH		0x48					// Switches A & C together
#define ESC_DOWN_SWITCH		0x90					// Switches B & D together
#define SWITCH_MASK (ENTER_SWITCH | ESC_SWITCH | UP_SWITCH | DOWN_SWITCH)	// Mask for all switch bits

typedef struct KEY_MAP_STRUCT						// Structure to hold mapping of scan codes to ASCII key values
{
	char scan_code;
	char key_code;
};


static const struct KEY_MAP_STRUCT key_map[] = 		// Maps key codes to ASCII Characters
{
	{UP_SWITCH, 'U'},								
	{DOWN_SWITCH, 'D'},																
	{ENTER_SWITCH, 'E'},
	{ESC_SWITCH, 'A'},
	{ENTER_ESC_SWITCH, 'W'},
	{UP_DOWN_SWITCH, 'X'},
	{ENTER_UP_SWITCH, 'Y'},
	{ESC_DOWN_SWITCH, 'Z'},
	{SWITCH_MASK, 0 },
};


unsigned char Count1ms = 10;						// Counts 100 uS tick for 1ms counter
unsigned char Tick1ms;								// Set by 1 ms background time function
char KeyDebounceTimer;								// Used to debounce keys
char TextBuffers[2][TEXT_BUFFER_LENGTH+1];			// Holds data to be displayed on LCD
bank1 char ReceiveBuffer[RECEIVE_BUFFER_LENGTH];	// Holds Data received from Controller
char ReceiveIndex;									// Index into recive buffer for receiving data	
char StxReceived;									// TRUE if the Start-of-Transmission character has been received (ESC)
char EtxReceived;									// TRUE if the End-of-Transmission character has been received (Zero)

char ContrastDelay, ContrastCount;					// Used to control the contrast using software PWM
char BacklightCount;								// Used to control backlight level using hardware PWM

// Function prototypes
void InitializeHardware(void);
void ProcessMessage(void);
void ProcessSerialInput(void);
void ClearTextBuffer(char Index);
void InitializeKeypad(void);
char GetKeyCode(unsigned char scan_code);
void DebounceKeypad(unsigned char keypad_bits);


// Timer overflows every 100 microseconds
static void interrupt timer_interrupt(void)
{
	if (TMR1IF)
	{
		if (--ContrastCount <= 0)
		{
			RB0 = !RB0;			
			ContrastCount = (RB0 == 1) ? ContrastDelay : (MAX_CONTRAST + 1) - ContrastDelay;
		}
		
		TMR1ON = 0;				// Have to turn off timer when changing count
		TMR1L = 0xB4;
		TMR1H = 0xFF;
		TMR1ON = 1;
		
		if (--Count1ms == 0)	// Set 1ms tick flag is 1ms has expired	
		{		
			Tick1ms = TRUE;
			Count1ms = 10;
		}
		TMR1IF = 0;				// Reset Timer flag
	}
}

void main(void)
{	
	InitializeHardware();
	InitializeKeypad();
	lcd_init();
	ClearTextBuffer(0);
	ClearTextBuffer(1);
	
	StxReceived = FALSE;		// Set up to receive serial data
	ReceiveIndex = 0;
	
	ContrastCount = ContrastDelay = MAX_CONTRAST / 2;		// set contrast to mid-range
	
										
	LCDGoto(0,0);								// Display some text with the version number
	LCDPutROMString(VER_STRING);	
	LCDPutROMString("ABCDEFGH");
	LCDGoto(1,0);	
	LCDPutROMString("0123456789012345");	
	
	// Main program loop - stays here forever	
	while(1)
	{
		asm("clrwdt");								// Clear the watchdog timer
				
		if (Tick1ms)								// Perform 1 millisecond tasks
		{
			if (--KeyDebounceTimer <= 0)			// Debounce keypad every ten milliseconds
			{
				DebounceKeypad(PORTA);	
				KeyDebounceTimer = DEBOUNCE_MSEC;
			}
			Tick1ms = FALSE;	
		}			
	
		if (key != NO_KEY)							// If we have a valid key...
		{
			SIO_XMIT(key);							// send it
			key = NO_KEY;
		}

		ProcessSerialInput();						// Get any data from the SIO and process it

													
		if (StxReceived == TRUE && EtxReceived == TRUE)// If message received, process it
		{
			ProcessMessage();						
			StxReceived = FALSE;
			EtxReceived = FALSE;
		}
	}		
}


// Delays ' Count' milliseconds, using 100us interrupt
// This function 'blocks' until the delay time has expired
void DelayMs(unsigned char Count)
{
	unsigned char Timer;
	
	GIE = 0;			// Disable interrupts while changing timer value
	Count1ms = 10;		// Reset the background timer value
	Tick1ms = FALSE;	// Reset the 1 millisecond tick counter
	GIE = 1;	
						// Loop 'Count' milliseconds
	for (Timer = Count; Timer; Timer--)
	{
		while (Tick1ms == FALSE);	// Wait for 1 mS to expire
		Tick1ms = FALSE;			// reset flag
	}
}	



void InitializeHardware(void)
{
	TRISB = 0;
	TRISA = 0xff;		// Sets switch bits to input
	CMCON = 0x07;
	VRCON = 0;
	RB3 = 1;
	RB0 = 0;
	
	T1CON = 0x01; 			// T1 is enabled, 1/1 prescaler, 
	TMR1L = 0xB4;
	TMR1H = 0xFF;
	TMR1IE = 1;
	PEIE = 1;
	GIE = 1;
	
	// Set up PWM on PORT B, bit 3 using Timer 2 as PWM to drive backlight level
	PR2 = 249;			// 
	BacklightCount = CCPR1L = 128;		// 50% duty Cycle
	CCP1CON = 0x0C;		// Set PWM Mode 
	T2CON = 0x05;
	
	// Setup Serial Port
	TXSTA = 0x20;	// clock from BRG, 8 bit tx, tx en, async, BRGH=0. 
	RCSTA = 0x90;   // enable sp, TX and RX 
	SPBRG = 51;		// @ 4MHz, 51 = 1200 Baud
	TRISB |= 0x06;	// Set both RX and TX pins as Input 
}


// Get a single character from the SIO and process it
// Builds message in 'ReceiveBuffer' - EtxReceived = TRUE on end-of-message 
void ProcessSerialInput(void)
{
	char rec_char;	
	
	if (SIO_REC_READY)						// Character in Receive Buffer
	{
		rec_char = SIO_GET;					// Get it
	
		if (rec_char == STX_CHAR)			// Check if start of message
		{
			StxReceived = TRUE;
			EtxReceived = FALSE;
			ReceiveIndex = 0;
		}
			
		else if (rec_char == ETX_CHAR)		// Check if end of message
			EtxReceived = TRUE;

		if (StxReceived)
		{
			ReceiveBuffer[ReceiveIndex] = rec_char;
			if (++ReceiveIndex >= RECEIVE_BUFFER_LENGTH)
			{
				ReceiveBuffer[ReceiveIndex -1] = 0;
				EtxReceived = TRUE;
			}
		}
	}	
}	


// Process the message in the recieve buffer
void ProcessMessage(void)
{
	char StringLength, Character;
	const char *SettingsPtr;
	
	if (ReceiveBuffer[0] == STX_CHAR)				// Make sure its a valid message
	{
		switch (ReceiveBuffer[1])					// Second character is message type
		{
			case 1:									// Text  for line 1
				StringLength = strlen(&ReceiveBuffer[2]);
				if (StringLength > TEXT_BUFFER_LENGTH)
					StringLength = TEXT_BUFFER_LENGTH;
					
				ClearTextBuffer(0);
				memcpy(TextBuffers[0], &ReceiveBuffer[2], StringLength);				
				break;
				
			case 2:									// Text for line 2				
				StringLength = strlen(&ReceiveBuffer[2]);
				if (StringLength > TEXT_BUFFER_LENGTH)
					StringLength = TEXT_BUFFER_LENGTH;
					
				ClearTextBuffer(1);
				memcpy(TextBuffers[1], &ReceiveBuffer[2], StringLength);
				break;
				
			case 'D':								// Display Control Command
				if (ReceiveBuffer[2] <= 0x30)		// Set backlight based on first data byte
					CCPR1L = 0;
				else if (ReceiveBuffer[2] >= 0x4f)
					CCPR1L = 255;
				else 
					CCPR1L = (ReceiveBuffer[2] - 0x30) << 3;


				if (ReceiveBuffer[3] <= 0x30)		// Set Contrast based on first data byte
					ContrastDelay = 0;
				else if (ReceiveBuffer[3] >= 0x4f)
					ContrastDelay = MAX_CONTRAST;
				else 
					ContrastDelay = (ReceiveBuffer[3] - 0x30) / 2;
				
				
				if (ReceiveBuffer[4] == '2')		// Set display On or Off
				{
					LCDOff();				
					CCPR1L = 0;						// If display off, turn off backlight
				}
				else 
					LCDOn();
				break;
				
				
			case 'C':								// Text Display Command
				if (ReceiveBuffer[2] == '0')		// Clear Display command
				{
					LCDClear();
					ClearTextBuffer(0);
					ClearTextBuffer(1);
				} 
			
				else if (ReceiveBuffer[2] == '1')	// Copy text from buffers to display	
				{
					LCDGoto(0,0);
					LCDPutString(TextBuffers[0]);
					LCDGoto(1,0);
					LCDPutString(TextBuffers[1]);
				}
				break;			
				
			case 'V':								// Return Module Settings Info
				SettingsPtr = SETTINGS_STRING;
				do
				{
					Character = *(SettingsPtr++);
					SIO_XMIT(Character);	
				} while (Character != 0);
				break;
		} 
	}
}


void ClearTextBuffer(char Index)
{
	memset(TextBuffers[Index], ' ', TEXT_BUFFER_LENGTH);
}

// Debounces the keypad and stores the resulting key code in the "key" variable
// This function must be called on a regular periodic basis - about once every 10 milliseconds
// The variable "keypad_bits" contains the data from the keypad.
void DebounceKeypad(unsigned char keypad_bits)
{
	unsigned char loop;
	char debounced;
	char new_key;
	static char last_key;
	static int repeat_delay, repeat_rate;	
	
	keypad_bits&= SWITCH_MASK;
	
	// Initially set the debounced state dependent on weather the new reading matches the last
	debounced = (keypad_bits == keypad_data[0]) ? TRUE : FALSE;			
	
	//Shift the old data and check if the last 'n' reading are the same
	for (loop = DEBOUNCE_CYCLES - 1; loop > 0; loop--)
	{
		if (keypad_data[loop] != keypad_data[loop - 1])	// Not debounced if consequitive readings are differant
			debounced = FALSE;
		keypad_data[loop] = keypad_data[loop - 1];		// Shift our keypad data through the history array
	}
	keypad_data[0] = keypad_bits;
	
	if (debounced == TRUE)
	{	// Get keycode corrsponding to key(s) pressed
		new_key = GetKeyCode(keypad_bits);

		// Perform auto repeat on UP & DOWN ARROW Keys only 
		// After key first pressed delay "DELAY_COUNT"
		// then repeat at "REPEAT_RATE"
		if (new_key == last_key)
		{
			if (new_key == 'U' || new_key == 'D')
			{
				if (repeat_delay == 0)				// If we've delayed enough from first key down...
				{
					if (repeat_rate == 0)			// If it's time to do a key...
					{			
						key = new_key;					// force a new key press
						repeat_rate = RATE_COUNT;	// and reset the rate time
					}
					else
						repeat_rate--;					// Not time for another key,  decrement counter				
				}
				else										// repeat_delay not Zero
				{
					repeat_delay--;					// Decrement auto-repeat delay
					repeat_rate = 0;					// Zero rate counter
				}
			}
		}
		
		else	// For other keys, we only return a valid key press on the key-down event
		{
			key = new_key;
			repeat_delay = DELAY_COUNT;
		}
		last_key = new_key;
	} // if (debounced == TRUE)
}

// Returns ASCII key value based on "scan code" (keypad port value)
char GetKeyCode(unsigned char scan_code)
{
	unsigned char index;	
	unsigned char key_code = NO_KEY;

	scan_code = ~scan_code;
	scan_code&= SWITCH_MASK;
	for (index = 0; index < Dim(key_map); index ++)
	{
		if (key_map[index].scan_code == scan_code)
		{
			key_code =  key_map[index].key_code;
			break;
		}
	}
	return(key_code);
}

void InitializeKeypad(void)
{
	unsigned char loop;
	
	key = NO_KEY;			// Set key to "No Key Pressed"
	for (loop = 0; loop < DEBOUNCE_CYCLES; loop++)	// Fill the keypad history array to inidctae noi key has been pressed 
		keypad_data[loop] = 0;
}
