
#ifndef __LCD_DRIVE_H
#define __LCD_DRIVE_H

// LCD peripheral routines.
//
//   Notes:
//      - These libraries routines are written to support the  Hitachi HD44780 LCD controller.
//      - The user must define the following items:
//          - The LCD interface type (4- or 8-bits)
//          - If 4-bit mode
//              - whether using the upper or lower nibble
//          - The data port
//              - The tris register for data port
//              - The control signal ports and pins
//              - The control signal port tris and pins
//          - The user must provide three delay routines:
//              - DelayFor18TCY() provides a 18 Tcy delay
//              - DelayPORLCD() provides at least 15ms delay
//              - DelayLCD() provides at least 5ms delay
//

// Interface type 8-bit or 4-bit -  For 8-bit operation uncomment the #define BIT8
// #define BIT

// When in 4-bit interface define if the data is in the upper or lower nibble.  
// For lower nibble, comment the #define UPPER
 #define UPPER

// DATA_PORT defines the port to which the LCD data lines are connected 
#define DATA_PORT      PORTB
#define TRIS_DATA_PORT TRISB

// Set your LCD type.
#define LCD_TYPE           (FOUR_BIT & LINES_5X8)

// This is how LCD will be setup on Init.
#define LCD_DISPLAY_SETUP  (CURSOR_OFF & BLINK_OFF)


// These define the bits used for the control lines
#define RW_PIN   RA2		   	// PORT for RW 
#define TRIS_RW  TRISA2 		// TRIS for RW 
#define RS_PIN   RA0		   	// PORT for RS
#define TRIS_RS  TRISA0    		// TRIS for RS
#define E_PIN    RA1   			// PORT for E
#define TRIS_E   TRISA1    		// TRIS for E

// Display ON/OFF Control defines 
#define DON         0b00001111  // Display on      
#define DOFF        0b00001011  // Display off     
#define CURSOR_ON   0b00001111  // Cursor on       
#define CURSOR_OFF  0b00001101  // Cursor off      
#define BLINK_ON    0b00001111  // Cursor Blink    
#define BLINK_OFF   0b00001110  // Cursor No Blink 

// Cursor or Display Shift defines 
#define SHIFT_CUR_LEFT    0b00010011  // Cursor shifts to the left   
#define SHIFT_CUR_RIGHT   0b00010111  // Cursor shifts to the right  
#define SHIFT_DISP_LEFT   0b00011011  // Display shifts to the left  
#define SHIFT_DISP_RIGHT  0b00011111  // Display shifts to the right 
#define DISP_CLEAR        0b00000001

// Function Set defines 
#define FOUR_BIT   0b00101111  // 4-bit Interface               
#define EIGHT_BIT  0b00111111  // 8-bit Interface               
#define LINE_5X7   0b00110011  // 5x7 characters, single line   
#define LINE_5X10  0b00110111  // 5x10 characters               
#define LINES_5X7  0b00111111  // 5x7 characters, multiple line 
#define LINES_5X8  0b00111000  // 5x8 characters, multiple line 


void lcd_init(void);
void LCDPut(char data);
void LCDPutString(char *string);
void LCDPutROMString(const char *string);
char LCDIsBusy(void);
void LCDCommand(unsigned char cmd);

#define LCDGoto(row, col)  	LCDCommand((col + (row * 0x40)) | 0x80)
#define LCDClear()         	LCDCommand(DISP_CLEAR)
#define LCDCursorRight()   	LCDCommand(SHIFT_CUR_RIGHT)
#define LCDCursorLeft()    	LCDCommand(SHIFT_CUR_LEFT)
#define LCDShiftLeft()     	LCDCommand(SHIFT_DISP_LEFT)
#define LCDShiftRight()    	LCDCommand(SHIFT_DISP_RIGHT)
#define LCDOn()				LCDCommand(DON);
#define LCDOff()			LCDCommand(DOFF);

#define LCDCGRAMAddr(addr) LCDCommand(addr | 0b01000000)
#define LCDDDRAMAddr(addr) LCDCommand(addr | 0b10000000)


void Delay50Us(unsigned char x);
void DelayMs(unsigned char cnt);

#endif
